#include "precompile.h"

#include "mt/Grasp.h"
#include "mt/Item.h"

IMPL_TABLE(mt::Grasp)
std::map<long long, std::tuple<int, std::set<int>>> mt::Grasp::hero_id_lv_gun_hash_;

namespace mt
{

    void Grasp::Init1()
    {
        {
            std::vector<std::string> strings;
            a8::Split(add_buff_list(), strings, '|');
            for (auto& str : strings) {
                _add_buff_list.insert(a8::XValue(str));
            }
        }
        {
            std::vector<std::string> strings;
            a8::Split(remove_buff_list(), strings, '|');
            for (auto& str : strings) {
                _remove_buff_list.insert(a8::XValue(str));
            }
        }
        if (hero_id() == 0) {
            abort();
        }
        if (hero_lv() == 0) {
            abort();
        }
        if (weapon_id() == 0) {
            abort();
        }
        const mt::Item* item_meta = mt::Item::GetById(weapon_id());
        if (!item_meta){
            abort();
        }
        weapon_id_ = item_meta->relationship();
    }

    void Grasp::Init2()
    {

    }

    void Grasp::StaticPostInit()
    {
        int last_hero_id = 0;
        int last_hero_lv = 0;
        int last_weapon_id = 0;
        std::set<int> last_buff_list;
        Grasp::MutTraverse
            (
             [&last_hero_id, &last_hero_lv, &last_weapon_id, &last_buff_list]
             (mt::Grasp* meta, bool& stop) mutable
             {
                 #ifdef DEBUG
                 a8::XPrintf("last_hero_id:%d last_hero_lv:%d\n", {last_hero_id, last_hero_lv});
                 #endif
                 long long key = a8::MakeInt64(meta->hero_id(), meta->hero_lv());
                 if (!last_hero_id) {
                     last_hero_id = meta->hero_id();
                     last_hero_lv = meta->hero_lv();
                     last_weapon_id = meta->weapon_id();
                     if (meta->hero_lv() != 1) {
                         abort();
                     }
                     auto itr = hero_id_lv_gun_hash_.find(key);
                     if (itr != hero_id_lv_gun_hash_.end()) {
                         abort();
                     }
                     auto tuple = std::make_tuple(meta->weapon_id(), std::set<int>());
                     for (int buff_id : meta->_add_buff_list) {
                         std::get<1>(tuple).insert(buff_id);
                         last_buff_list.insert(buff_id);
                     }
                     for (int buff_id : meta->_remove_buff_list) {
                         std::get<1>(tuple).erase(buff_id);
                         last_buff_list.erase(buff_id);
                     }
                     hero_id_lv_gun_hash_[key] = tuple;
                 } else {
                     if (5 * (last_hero_lv / 5 + 1) != meta->hero_lv()) {
                         abort();
                     }
                     if (last_weapon_id != meta->weapon_id()) {
                         abort();
                     }
                     last_hero_id = meta->hero_id();
                     last_hero_lv = meta->hero_lv();
                     last_weapon_id = meta->weapon_id();

                     auto tuple = std::make_tuple(meta->weapon_id(), last_buff_list);
                     for (int buff_id : meta->_add_buff_list) {
                         std::get<1>(tuple).insert(buff_id);
                         last_buff_list.insert(buff_id);
                     }
                     for (int buff_id : meta->_remove_buff_list) {
                         std::get<1>(tuple).erase(buff_id);
                         last_buff_list.erase(buff_id);
                     }
                     hero_id_lv_gun_hash_[key] = tuple;

                     if (last_hero_lv == 15) {
                         last_hero_id = 0;
                         last_hero_lv = 0;
                         last_weapon_id = 0;
                         last_buff_list.clear();
                     }
                 }
             });
    }

    std::set<int>* Grasp::GetBuffs(int hero_id, int hero_lv, int gun_id)
    {
        int internal_lv = 1;
        switch (hero_lv) {
        case 1:
        case 2:
        case 3:
        case 4:
            {
                internal_lv = 1;
            }
            break;
        case 5:
        case 6:
        case 7:
        case 8:
        case 9:
            {
                internal_lv = 5;
            }
            break;
        case 10:
        case 11:
        case 12:
        case 13:
        case 14:
            {
                internal_lv = 10;
            }
            break;
        default:
            {
                if (hero_lv < 1) {
                    internal_lv = 1;
                } else {
                    internal_lv = 15;
                }
            }
            break;
        }
        long long key = a8::MakeInt64(hero_id, internal_lv);
        auto itr = hero_id_lv_gun_hash_.find(key);
        if (itr != hero_id_lv_gun_hash_.end()) {
            if (std::get<0>(itr->second) == gun_id) {
                return &std::get<1>(itr->second);
            }
        }
        return nullptr;
    }

}
